# coding: utf-8
# +-------------------------------------------------------------------
# | 宝塔Linux面板
# +-------------------------------------------------------------------
# | Copyright (c) 2015-2017 宝塔软件(http://bt.cn) All rights reserved.
# +-------------------------------------------------------------------
# | Author: hwliang <hwl@bt.cn>
# | Author: 1249648969@qq.com
# docker管理插件
# +--------------------------------------------------------------------
import sys, os, re
import time
import json
import docker

if sys.version_info[0] == 2:
    reload(sys)
    sys.setdefaultencoding('utf-8')
os.chdir('/www/server/panel')
sys.path.append("class/")
import public


class docker_main(object):
    __docker = None
    __path = '/www/backup/docker/'
    __setupPath = '/www/server/panel/plugin/docker'

    # 实例化SDK
    def __init__(self):
        if not os.path.exists(self.__path):
            public.ExecShell('mkdir -p %s' % self.__path)
        try:
            if not self.__docker: self.__docker = docker.from_env()
        except:
            pass

    # 判断服务是否正常
    def GetStatus(self, get):
        sock = '/var/run/docker.pid'
        if os.path.exists(sock):
            try:
                __docker = docker.from_env()
                aaa = __docker.images
                if aaa:
                    return public.returnMsg(True, 'Docker service is Running')
                else:
                    return public.returnMsg(False, 'Docker service is not started')
            except:
                return public.returnMsg(False, 'Docker service is not started')
        else:
            return public.returnMsg(False, 'Docker service is not started')

    # 取容器列表
    def GetConList(self, get):
        conList = []
        for con in self.__docker.containers.list(all=True):
            tmp = con.attrs
            tmp['Created'] = self.utc_to_local(tmp['Created'].split('.')[0])
            conList.append(tmp)
        return conList

    # UTC时间转换为时间戳
    def utc_to_local(self, utc_time_str, utc_format='%Y-%m-%dT%H:%M:%S'):
        import pytz, datetime, time
        local_tz = pytz.timezone('Asia/Chongqing')
        local_format = "%Y-%m-%d %H:%M"
        utc_dt = datetime.datetime.strptime(utc_time_str, utc_format)
        local_dt = utc_dt.replace(tzinfo=pytz.utc).astimezone(local_tz)
        time_str = local_dt.strftime(local_format)
        return int(time.mktime(time.strptime(time_str, local_format)))

    # 创建容器
    def CreateCon(self, get):
        try:
            environments = get.environments.strip().split() if 'environments' in get else []
            if not get.command: get.command = None
            if not get.entrypoint: get.entrypoint = None
            conObject = self.__docker.containers.run(
                image=get.image,
                mem_limit=get.mem_limit + 'M',
                ports=eval(get.ports.replace('[', '(').replace(']', ')')),
                auto_remove=False,
                command=get.command,
                detach=True,
                stdin_open=True,
                tty=True,
                entrypoint=get.entrypoint,
                privileged=True,
                volumes=json.loads(get.volumes),
                cpu_shares=10,
                environment=environments
            )
            if conObject:
                # conObject.exec_run('(echo "'+get.ps.strip()+'";sleep 0.1;echo "'+get.ps.strip()+'")|passwd root',privileged = True,stdin = True);
                self.AcceptPort(get)
                return public.returnMsg(True, 'Created successfully!')

            return public.returnMsg(False, 'Created failed!')
        except docker.errors.APIError as ex:
            return public.returnMsg(False, 'Created failed!' + str(ex))

    # pull Dockr 官方镜像
    def pull(self, get):
        images = get.images
        if ':' in images:
            pass
        else:
            images = images + ':latest'
        try:
            ret = self.__docker.images.pull(images)
            if ret:
                return public.returnMsg(True, 'Pull success!')
            else:
                return public.returnMsg(False, 'Pull failed, please check the name of the image or whether you need to log in to docker to download')
        except:
            ret = public.ExecShell('docker image pull %s' % images)
            if 'invalid' in ret[-1]:
                return public.returnMsg(False, 'Pull failed, please check the name of the image or whether you need to log in to docker to download')
            else:
                return public.returnMsg(True, 'Pull success!')

    # pull 其他
    def pull_reg(self, get):
        ret = 'not\s?found\n$|invalid|Error'
        path = str(get.path)
        if not path and path == '': return public.returnMsg(False, 'Invalid address')
        ret2 = public.ExecShell('docker image pull %s' % path)
        retur = re.findall(ret, ret2[-1])
        if len(retur) == 0:
            return public.returnMsg(True, 'Pull success!')
        else:
            return public.returnMsg(False, 'Invalid address')

    # pull 私有仓库
    def pull_private(self, get):
        user_name = get.user
        user_pass = get.password
        registry = get.registry
        path = get.path
        if not user_name and not user_pass and not path: return public.returnMsg(False, 'Please enter the correct information')
        login = self.login_check(user_name=user_name, user_pass=user_pass, registry=registry)
        if login:
            ret = self.pull_reg(get)
            return ret
        else:
            return public.returnMsg(False, 'Login failed')

    def my_repository(self, get):
        repostory_info = []
        user_file = '/www/server/panel/plugin/docker/user.json'
        user_info = public.readFile(user_file)
        if user_info:
            user_info = json.loads(user_info)
            for i in user_info:
                tmp = {}
                tmp["hub_name"] = i["hub_name"]
                tmp["registry"] = i["registry"]
                tmp["namespace"] = i["namespace"]
                tmp['repository_name'] = i["repository_name"]
                repostory_info.append(tmp)
        return repostory_info

    # 判断镜像是否存在
    def check_image(self, get):
        image_list = self.GetImageList(get)
        for i in image_list:
            if get.path == i["RepoTags"]:
                return public.returnMsg(False, 'The image already exists')

    def pull_private_new(self, get):
        """
        get.path  pull地址
        :param get:
        :return:
        """
        result = self.check_image(get)
        if result:
            return result
        my_repository = self.my_repository(get)
        if not my_repository:
            return public.returnMsg(False, 'Did not log in any private repository, please log in and then pull')
        result = self.pull_reg(get)
        return result

    # 放行端口
    def AcceptPort(self, get):
        import firewalls
        f = firewalls.firewalls()
        get.ps = 'Port of docker container'
        for port in json.loads(get.accept):
            get.port = port
            f.AddAcceptPort(get)
        return True

    # 判断端口是否被占用
    def IsPortExists(self, get):
        port = get.port
        ret = self.__check_dst_port(ip='localhost', port=port)
        ret2 = self.__check_dst_port(ip='0.0.0.0', port=port)
        if ret:
            return ret
        if not ret and ret2:
            return ret2
        if not ret and not ret2:
            return False

    # 端口检测
    def __check_dst_port(self, ip, port, timeout=3):
        import socket
        ok = True
        try:
            s = socket.socket()
            s.settimeout(timeout)
            s.connect((ip, port))
            s.close()
        except:
            ok = False
        return ok

    # 删除容器
    def RemoveCon(self, get):
        try:
            conFind = self.__docker.containers.get(get.Hostname)
            try:
                path_list = conFind.attrs['GraphDriver']['Data']['LowerDir'].split(':')
                for i in path_list:
                    public.ExecShell('chattr -R -i %s' % i)
            except:
                pass
            conFind.remove(force=True)
            return public.returnMsg(True, 'Deleted successfully!')
        except docker.errors.APIError as ex:
            return public.returnMsg(False, 'Failed to delete!' + str(ex))

    # 启动容器
    def RunCon(self, get):
        try:
            conFind = self.__docker.containers.get(get.Hostname)
            if not conFind: return public.returnMsg(False, 'The specified container does not exist!')
            conFind.start()
            return public.returnMsg(True, 'Startup successful!')
        except docker.errors.APIError as ex:
            return public.returnMsg(False, 'Startup failed!' + str(ex))

    # 停止容器
    def StopCon(self, get):
        try:
            conFind = self.__docker.containers.get(get.Hostname)
            if not conFind: return public.returnMsg(False, 'The specified container does not exist!')
            conFind.stop()
            return public.returnMsg(True, 'Stop successful!')
        except docker.errors.APIError as ex:
            return public.returnMsg(False, 'Stop failed!' + str(ex))

    # 修改名称
    def ReName(self, get):
        try:
            conFind = self.__docker.containers.get(get.Hostname)
            if not conFind: return public.returnMsg(False, 'The specified container does not exist!')
            conFind.rename(get.name2)
            return public.returnMsg(True, 'Modified successfully!')
        except docker.errors.APIError as ex:
            return public.returnMsg(False, 'Modification failed! Duplicate name')

    # 获取日志
    def get_log(self, get):
        try:
            conFind = self.__docker.containers.get(get.Hostname)
            if not conFind: return public.returnMsg(False, 'The specified container does not exist!')
            log = conFind.logs()
            if not isinstance(log, str):
                log = log.decode()
            return public.returnMsg(True, log)
        except docker.errors.APIError as ex:
            return public.returnMsg(False, 'Get Logs failed')

    # 导出容器为tar 格式
    def export(self, get):
        try:
            file_name = self.__path + get.Hostname + str(time.strftime('%Y%m%d_%H%M%S', time.localtime())) + '.tar'
            public.ExecShell('mkdir -p %s && touch %s' % (self.__path, file_name))
            f = open(file_name, 'wb')
            conFind = self.__docker.containers.get(get.Hostname)
            data = conFind.export()
            for i in data.read():
                f.write(i)
            f.close()
            return public.returnMsg(True, file_name)
        except docker.errors.APIError as ex:
            return public.returnMsg(False, 'Operation failed: ' + str(ex))

    # image 导出
    def save_image(self, get):
        try:
            file_name = self.__path + str(time.strftime('%Y%m%d_%H%M%S', time.localtime())) + '.tar.gz'
            public.ExecShell('docker image save %s | gzip > %s' % (get.images, file_name))
            return public.returnMsg(True, 'Export image to {} successful!'.format(file_name))
        except docker.errors.APIError as ex:
            return public.returnMsg(False, 'Operation failed: ' + str(ex))

    # image 导入
    def load_image(self, get):
        try:
            file_path = get.file_path.strip()
            if not os.path.exists(file_path):
                return public.returnMsg(False, 'File does not exist')
            if file_path.endswith('.tar'):
                public.ExecShell('docker image load < %s' % file_path)
            elif file_path.endswith('.tar.gz'):
                public.ExecShell('gunzip -c %s | docker image load' % file_path)
            else:
                return public.returnMsg(False, 'Unsupported file type')
            return public.returnMsg(True, 'Import image successfully!')
        except docker.errors.APIError as ex:
            return public.returnMsg(False, 'Operation failed: ' + str(ex))

    # 获取本地备份文件列表
    def get_backup_file_list(self, get):
        files = []
        for file_name in os.listdir(self.__path):
            file_path = os.path.join(self.__path, file_name)
            if not os.path.exists(file_path): continue
            f_info = {}
            f_info['file_path'] = file_path
            f_info['mtime'] = int(os.path.getmtime(file_path))
            f_info['size'] = public.get_path_size(file_path)
            files.append(f_info)
        files = sorted(files, key=lambda x: x['mtime'], reverse=True)
        return files

    # 容器执行命令
    def ExecutiveOrder(self, get):
        debug_path = 'data/debug.pl'
        if os.path.exists(debug_path):
            return public.returnMsg(False, 'The panel terminal cannot be opened in developer mode')
        try:
            conFind = self.__docker.containers.get(get.Hostname)
            cmd = 'docker container exec -it %s /bin/bash' % get.Hostname
            return public.returnMsg(True, cmd)
        except docker.errors.APIError as ex:
            return public.returnMsg(False, 'Connection failed')

    # 登陆验证
    def login_check(self, user_name, user_pass, registry):
        login_test = public.ExecShell('docker login -u=%s -p %s %s' % (user_name, user_pass, registry))
        ret = 'required$|Error'
        ret2 = re.findall(ret, login_test[-1])
        if len(ret2) == 0:
            return True
        else:
            return False

    # 登陆阿里云Docker hub
    def login(self, get):
        '''
        只支持阿里云 docker hub
        :param get:
        :return:
        '''
        repository_name = get.repository_name
        user_name = get.user
        user_pass = get.passwd
        registry = get.registry
        hub_name = get.hub_name
        namespace = get.namespace
        if not user_pass and not user_pass and not registry and not hub_name and not namespace: return public.returnMsg(
            False, 'Please fill in the correct information')
        ret2 = self.login_check(user_name, user_pass, registry)
        if ret2:
            ret = {}
            ret['user_name'] = user_name
            ret['user_pass'] = user_pass
            ret['registry'] = registry
            ret['hub_name'] = hub_name
            ret['namespace'] = namespace
            ret['repository_name'] = repository_name
            public.writeFile(self.__setupPath + '/user.json', json.dumps(ret))

            return public.returnMsg(True, 'Landed successfully')
        else:
            return public.returnMsg(False, 'Login failed')

    # 兼容旧用户文件
    def compatible_old_userinfo(self):
        user_info_list = []
        user_file = '/www/server/panel/plugin/docker/user.json'
        user_info = public.readFile(user_file)
        if user_info:
            user_info = json.loads(user_info)
            if isinstance(user_info, dict):
                user_info_list.append(user_info)
                public.writeFile(user_file, json.dumps(user_info_list))

    # 获取需要操作的仓库
    def get_operating_repository(self,repository_name):
        user_file = '/www/server/panel/plugin/docker/user.json'
        user_info = public.readFile(user_file)
        if user_info:
            user_info = json.loads(user_info)
            for i in user_info:
                if repository_name in i.values():
                    return i

    # 登陆私有DockerHub
    def login_private(self,get):
        '''
        get.user        用户名
        get.passwd      用户密码
        get.registry    仓库域名/地址(docker_hub 留空)
        get.hub_name    仓库名
        get.namespace   命名空间
        :param get:
        :return:
        '''
        user_name = get.user
        user_pass = get.passwd
        registry = get.registry
        hub_name = get.hub_name
        namespace = get.namespace
        repository_name = get.repository_name
        result = self.login_check(user_name, user_pass, registry)

        if result:
            user_file = self.__setupPath + '/user.json'
            self.compatible_old_userinfo()
            user_info = public.readFile(user_file)
            if not user_info:
                user_info = []
            else:
                user_info = json.loads(user_info)
            ret = {}
            ret['user_name'] = user_name
            ret['user_pass'] = user_pass
            ret['registry'] = registry
            ret['hub_name'] = hub_name
            ret['namespace'] = namespace
            ret['repository_name'] = repository_name
            if not registry:
                ret['registry'] = "docker.io"
            user_info.append(ret)
            public.writeFile(self.__setupPath + '/user.json', json.dumps(user_info))
            return public.returnMsg(True, 'Login successfully')
        return public.returnMsg(False, 'Login failed')

    # 删除用户信息
    def delete_user_info(self,registry):
        user_file = '/www/server/panel/plugin/docker/user.json'
        user_info = public.readFile(user_file)
        if user_info:
            user_info = json.loads(user_info)
            for i in range(len(user_info)):
                if registry in user_info[i].values():
                    del(user_info[i])
                    public.writeFile(user_file,json.dumps(user_info))
                    return True

    # 登出
    def logout_private(self,get):
        if get.registry == "docker.io":get.registry = ""
        login_test = public.ExecShell('docker logout %s' % get.registry)
        if get.registry == "": get.registry = "docker.io"
        ret = 'required$|Error'
        ret2 = re.findall(ret, login_test[-1])
        self.delete_user_info(get.registry)
        if len(ret2) == 0:
            return public.returnMsg(True, 'Logout success')
        else:
            return public.returnMsg(True, 'Logout failed')

    # push 到仓库中
    def push(self, get):
        version = get.version
        imageid = get.imageid
        repository_name = get.repository_name
        repository = self.get_operating_repository(repository_name)
        if not repository:
            return public.returnMsg(False, 'Please login')
        registry = self.chekc(repository['registry'])
        hub_name = self.chekc(repository['hub_name'])
        namespace = self.chekc(repository['namespace'])
        user_name = repository['user_name']
        user_pass = repository['user_pass']
        if not user_name and user_pass and not hub_name and not namespace: public.returnMsg(False,
                                                                                                     'Information verification failed')
        ret2 = self.login_check(repository["user_name"], repository["user_pass"], registry)
        if ret2:
            local_image = '%s/%s/%s:%s' % (registry, namespace, hub_name, version)
            public.ExecShell('docker image tag %s %s' % (imageid, local_image))
            push_result = public.ExecShell('docker image push %s' % local_image)
            self.__docker.images.remove(local_image)

            ret = 'exist'
            rec = re.findall(ret, push_result[-1])
            if len(rec) == 0:
                return public.returnMsg(True, 'Push Success')
            else:
                return public.returnMsg(False, 'Push failed')
        else:
            return public.returnMsg(False, 'Login failed')

    # 判断最后一个是否是/
    def chekc(self, string):
        strings = str(string)
        if strings == '': return False
        if strings[-1] == '/': return strings[0:-1]
        return strings

    # 重启容器
    def RestartCon(self, get):
        try:
            conFind = self.__docker.containers.get(get.Hostname)
            if not conFind: return public.returnMsg(False, 'The specified container does not exist!');
            conFind.restart()
            return public.returnMsg(True, 'Restart successfully!')
        except docker.errors.APIError as ex:
            return public.returnMsg(False, 'Restart successfully!' + str(ex))

    # 登陆状态
    def GetUser(self, get):
        relist = self.__setupPath + '/user.json'
        if os.path.exists(relist):
            user_info = json.loads(public.readFile(relist))
            if user_info['user_name']:
                return public.returnMsg(True, user_info)
            else:
                return public.returnMsg(False, 'Not login')
        else:
            return public.returnMsg(False, 'Not login')

    # 取创建依赖
    def GetCreateInfo(self, get):
        import psutil
        data = {}
        data['images'] = self.GetImageList(None)
        data['memSize'] = int(psutil.virtual_memory().total / 1024 / 1024)
        data['iplist'] = self.GetIPList(None)
        return data

    # 取IP列表
    def GetIPList(self, get):
        ipConf = '/www/server/panel/plugin/docker/iplist.json'
        if not os.path.exists(ipConf): return []
        iplist = json.loads(public.readFile(ipConf))
        return iplist

    # 添加IP
    def AddIP(self, get):
        ipConf = '/www/server/panel/plugin/docker/iplist.json'
        if not os.path.exists(ipConf):
            iplist = []
            public.writeFile(ipConf, json.dumps(iplist))

        iplist = json.loads(public.readFile(ipConf))
        ipInfo = {'address': get.address, 'netmask': get.netmask, 'gateway': get.gateway}
        iplist.append(ipInfo)
        public.writeFile(ipConf, json.dumps(iplist))
        return public.returnMsg(True, 'Added successfully!')

    # 删除IP
    def DelIP(self, get):
        ipConf = '/www/server/panel/plugin/docker/iplist.json'
        if not os.path.exists(ipConf): return public.returnMsg(False, 'The specified IP does not exist.!')
        iplist = json.loads(public.readFile(ipConf))
        newList = []
        for ipInfo in iplist:
            if ipInfo['address'] == get.address: continue
            newList.append(ipInfo)
        public.writeFile(ipConf, json.dumps(newList))
        return public.returnMsg(True, 'Successfully deleted!')

    # 制作镜像
    def CommitCon(self, get):
        try:
            conFind = self.__docker.containers.get(get.Hostname)
            conFind.commit(repository=get.imageName, tag=get.tag, message=get.message, author=get.author)
            return public.returnMsg(True, 'Operation successful!')
        except docker.errors.APIError as ex:
            return public.returnMsg(False, 'Operation failed: ' + str(ex))

    # 取镜像列表
    def GetImageList(self, get):
        imageList = []
        for image in self.__docker.images.list():
            tmp_attrs = image.attrs
            if len(tmp_attrs['RepoTags']) == 1:
                tmp_image = {}
                tmp_image['Id'] = tmp_attrs['Id'].split(':')[1][:12]
                tmp_image['RepoTags'] = tmp_attrs['RepoTags'][0]
                tmp_image['Size'] = tmp_attrs['Size']
                tmp_image['Labels'] = tmp_attrs['Config']['Labels']
                tmp_image['Comment'] = tmp_attrs['Comment']
                tmp_image['Created']=self.utc_to_local(tmp_attrs['Created'].split('.')[0])
                imageList.append(tmp_image)
            else:
                for i in range(len(tmp_attrs['RepoTags'])):
                    tmp_image = {}
                    tmp_image['Id'] = tmp_attrs['Id'].split(':')[1][:12]
                    tmp_image['RepoTags'] = tmp_attrs['RepoTags'][i]
                    tmp_image['Size'] = tmp_attrs['Size']
                    tmp_image['Labels'] = tmp_attrs['Config']['Labels']
                    tmp_image['Comment'] = tmp_attrs['Comment']
                    tmp_image['Created'] = self.utc_to_local(tmp_attrs['Created'].split('.')[0])
                    imageList.append(tmp_image)
        imageList = sorted(imageList, key=lambda x: x['Created'], reverse=True)
        return imageList

    # 删除镜像
    def RemoveImage(self, get):
        try:
            self.__docker.images.remove(get.repoTags)
            return public.returnMsg(True, 'Successfully deleted!')
        except:
            try:
                self.__docker.images.remove(get.imageId)
                return public.returnMsg(True, 'Successfully deleted!')
            except docker.errors.APIError as ex:
                return public.returnMsg(False, 'Delete failed, the image is currently being used!')

    # 用户管理页面
    def UserAdmin(self, get):
        return public.readFile('/www/server/panel/plugin/docker/userdocker.html')

    # 用户登陆
    def UserLogin(self, get):
        return public.readFile('/www/server/panel/plugin/docker/login-docker.html')

    # 获取镜像加速配置
    def get_registry_mirrors(self, get):
        try:
            if not os.path.exists('/etc/docker/daemon.json'):
                return public.returnMsg(True, [])
            conf = json.loads(public.readFile('/etc/docker/daemon.json'))
            return public.returnMsg(True, conf['registry-mirrors'])
        except:
            return public.returnMsg(False, 'Get failed! Reason for failure:{}'.format(public.get_error_info()))

    # 设置镜像加速配置
    def set_registry_mirrors(self, get):
        try:
            registry_mirrors = get.registry_mirrors.strip().split(',')
            conf = {}
            if os.path.exists('/etc/docker/daemon.json'):
                conf = json.loads(public.readFile('/etc/docker/daemon.json'))
            conf['registry-mirrors'] = registry_mirrors
            public.writeFile('/etc/docker/daemon.json', json.dumps(conf, indent=2))
            get.type = 'restart'
            self.service_admin(get)
            return public.returnMsg(True, 'Setup successfully')
        except:
            return public.returnMsg(False, 'Get failed! Reason for failure:{}'.format(public.get_error_info()))

    # docker服务启动，停止，重启
    def service_admin(self, get):
        s_type = get.type.strip().lower()
        if s_type not in ['start', 'stop', 'restart']:
            return public.returnMsg(False, 'Incorrect operation')

        exec_str = 'systemctl {} docker'.format(s_type)
        public.ExecShell(exec_str)
        if s_type in ['start', 'restart']:
            self.__docker = docker.from_env()
            for container in self.__docker.containers.list(all=True):
                try:
                    container.start()
                except:
                    pass
        return public.returnMsg(True, '{} success'.format(s_type))
